#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sndfile.h>
#ifdef _WIN32
#define NOMINMAX
#include <windows.h>
#endif
#include "softclip.h"
#include "clipper.h"

typedef struct file_t {
    SNDFILE *ifile, *ofile;
    SF_INFO iinfo, oinfo;
} file_t;

extern "C" SNDFILE* sf_wchar_open(const wchar_t*, int, SF_INFO*);

#ifdef _WIN32
SNDFILE* sf_openx(const char *path, int mode, SF_INFO *sfinfo)
{
    int size;
    wchar_t *wp;
    size = MultiByteToWideChar(65001, 0, path, -1, 0, 0);
    wp = (wchar_t*)_alloca(size * sizeof(wchar_t));
    size = MultiByteToWideChar(65001, 0, path, -1, wp, size);
    return sf_wchar_open(wp, mode, sfinfo);
}
#else
#define sf_openx sf_open
#endif

void usage(void)
{
    fprintf(stderr,
"usage: softclip [original|fixed] infile outfile\n"
"original: use original opus_pcm_soft_clip()\n"
"fixed:    use fixed clipper\n"
"both infile and outfile can be -, which means stdin/stdout\n"
    );
    exit(1);
}

SNDFILE *open_input(const char *file, SF_INFO *info)
{
    SNDFILE *fp;
    if (!strcmp(file, "-")) {
	if (!(fp = sf_open_fd(0, SFM_READ, info, 0)))
	    fputs("ERROR: sf_open_fd()\n", stderr);
    } else {
	if (!(fp = sf_openx(file, SFM_READ, info)))
	    fprintf(stderr, "ERROR: sf_open(): %s\n", file);
    }
    return fp;
}

SNDFILE *open_output(const char *file, SF_INFO *info)
{
    SNDFILE *fp;
    if (!file || !strcmp(file, "-")) {
	if (!(fp = sf_open_fd(1, SFM_WRITE, info, 0)))
	    fputs("ERROR: sf_open_fd()\n", stderr);
    } else {
	if (!(fp = sf_openx(file, SFM_WRITE, info)))
	    fprintf(stderr, "ERROR: sf_open(): %s\n", file);
    }
    return fp;
}

void process_file_opus(const file_t *file)
{
    const size_t NS = 0x1000;
    sf_count_t count = 0;
    size_t nch = file->iinfo.channels;
    std::vector<float> buf(NS * nch);
    std::vector<float> declip_mem(nch);
    float *bufp = &buf[0];

    while ((count = sf_readf_float(file->ifile, bufp, NS)) > 0) {
        opus_pcm_soft_clip(bufp, count, nch, &declip_mem[0]);
        sf_writef_float(file->ofile, bufp, count);
    }
}

void process_file(const file_t *file)
{
    const size_t NS = 0x1000;
    sf_count_t count = 0;
    size_t nch = file->iinfo.channels;
    std::vector<float> buf(NS * nch);
    float *bufp = &buf[0];
    SoftClipper<float> clipper(nch);
    size_t nout;

    while ((count = sf_readf_float(file->ifile, bufp, NS)) > 0) {
        nout = NS;
        clipper.process(bufp, count, bufp, &nout);
        if (nout > 0)
            sf_writef_float(file->ofile, bufp, nout);
    }
    nout = NS;
    clipper.process(&buf[0], 0, bufp, &nout);
    if (nout > 0)
        sf_writef_float(file->ofile, bufp, nout);
}

#ifdef _WIN32
extern "C" int utf8_main(int argc, char **argv)
#else
extern "C" int main(int argc, char **argv)
#endif
{
    file_t file = { 0 };
    int rc = 2;
    const char * mode = 0;

    if (argc < 4)
        usage();
    mode = argv[1];
    if (strcmp(mode, "original") && strcmp(mode, "fixed")) {
        fprintf(stderr, "mode has to be either \"original\" or \"fixed\"\n");
        return 1;
    }
    argv += 2;

    if (!(file.ifile = open_input(argv[0], &file.iinfo)))
	goto done;

    file.oinfo.format = SF_FORMAT_WAV | SF_FORMAT_FLOAT;
    file.oinfo.samplerate = file.iinfo.samplerate;
    file.oinfo.channels = file.iinfo.channels;

    if (!(file.ofile = open_output(argv[1], &file.oinfo)))
	goto done;

    if (strcmp(mode, "original") == 0)
        process_file_opus(&file);
    else
        process_file(&file);
    rc = 0;
done:
    if (file.ifile) sf_close(file.ifile);
    if (file.ofile) sf_close(file.ofile);
    return rc;
}
