#include <cstring>
#include <algorithm>
#include <cmath>
#include <vector>

#ifdef _MSC_VER
typedef intptr_t ssize_t;
#endif

template <typename T>
T clip(T x, T low, T high)
{
    return std::max(low, std::min(high, x));
}

template <typename T>
class SoftClipper {
    int m_nchannels;
    T m_peak;
    std::vector<std::vector<T> > m_buffer;
    std::vector<size_t> m_processed;
public:
    explicit SoftClipper(int nchannels, T peak=0.9921875)
        : m_nchannels(nchannels), m_peak(peak)
    {
        m_buffer.resize(nchannels);
        m_processed.resize(nchannels);
    }
    void process(const T *in, size_t nin, T *out, size_t *nout)
    {
        for (int n = 0; n < m_nchannels; ++n) {
            std::vector<T> &x = m_buffer[n];
            x.reserve(x.size() + nin);
            for (size_t i = 0; i < nin; ++i) {
                x.push_back(clip(in[i * m_nchannels + n],
                                 static_cast<T>(-2),
                                 static_cast<T>(2)));
            }
            ssize_t limit = x.size();
            if (limit > 0 && nin > 0) {
                T last = x[limit-1];
                for (; limit > 0 && x[limit-1] * last > 0; --limit)
                    ;
            }
            ssize_t end = m_processed[n];
            while (end < limit) {
                ssize_t peak_pos = end;
                for (; peak_pos < limit; ++peak_pos)
                    if (x[peak_pos] > m_peak || x[peak_pos] < -m_peak)
                        break;
                if (peak_pos == limit)
                    break;
                ssize_t start = peak_pos;
                T peak = std::abs(x[peak_pos]);

                while (start > end && x[peak_pos] * x[start] >= 0)
                    --start;
                for (end = peak_pos + 1; end < limit; ++end) {
                    if (x[peak_pos] * x[end] < 0)
                        break;
                    T y = std::abs(x[end]);
                    if (y > peak) {
                        peak = y;
                        peak_pos = end;
                    }
                }
                T a = (peak - m_peak) / (peak * peak);
                if (x[peak_pos] > 0)
                    a = -a;
                for (ssize_t i = start; i < end; ++i)
                    x[i] = x[i] + a * x[i] * x[i];
            }
            m_processed[n] = limit;
        }
        size_t prod = std::min(*nout, *std::min_element(m_processed.begin(),
                                                        m_processed.end()));
        size_t k = 0;
        for (size_t i = 0; i < prod; ++i)
            for (int n = 0; n < m_nchannels; ++n)
                out[k++] = m_buffer[n][i];

        for (int n = 0; n < m_nchannels; ++n) {
            std::vector<T> &x = m_buffer[n];
            if (x.size() && prod) {
                std::rotate(x.begin(), x.begin() + prod, x.end());
                x.resize(x.size() - prod);
                m_processed[n] -= prod;
            }
        }
        *nout = prod;
    }
};
